@@ -35,6 +35,11 @@ CONCATENATOR_DATA_FILES = (
     "https://packages.texta.ee/texta-resources/concatenator/space_between_not_ok.txt",
 )
 
+# URLs for Custom NER model downloads.
+CUSTOM_NER_MODELS = {
+    "et": "https://packages.texta.ee/texta-resources/ner_models/_estonian_nertagger.pt",
+}
+
 # Location of the resource dir where models are downloaded
 DEFAULT_RESOURCE_DIR = os.getenv("TEXTA_MLP_DATA_DIR", os.path.join(os.getcwd(), "data"))
 
@@ -51,17 +56,32 @@ SUPPORTED_ANALYZERS = (
     "emails",
     "phone_strict",
     "entities",
-    "namemail",
-    "bounded",
     "currency_sum",
     "sentences"
 )
 
+DEFAULT_ANALYZERS = [
+    "lemmas",
+    "pos_tags",
+    "transliteration",
+    "ner",
+    "addresses",
+    "emails",
+    "phone_strict",
+    "entities",
+    "sentences",
+    "currency_sum"
+]
+
+
 # Here we define languages with NER support to avoid Stanza trying to load them for languages without NER support.
 # This significantly increases performance for languages without NER.
 # https://stanfordnlp.github.io/stanza/available_models.html#available-ner-models
 STANZA_NER_SUPPORT = ("ar", "zh", "nl", "en", "fr", "de", "ru", "es", "uk")
 
+# Here we add langs that will have custom ner models.
+CUSTOM_NER_MODEL_LANGS = ["et"]
+
 
 class MLP:
 
@@ -71,6 +91,7 @@ class MLP:
             default_language_code=DEFAULT_LANG_CODES[0],
             use_default_language_code=True,
             resource_dir: str = DEFAULT_RESOURCE_DIR,
+            ner_model_langs: list = CUSTOM_NER_MODEL_LANGS,
             logging_level="error",
             use_gpu=True,
             refresh_data=REFRESH_DATA
@@ -82,28 +103,17 @@ class MLP:
         self.resource_dir = resource_dir
 
         self._stanza_pipelines = {}
+        self.custom_ner_model_langs = ner_model_langs
         self.logging_level = logging_level
         self.use_gpu = use_gpu
         self.stanza_resource_path = pathlib.Path(self.resource_dir) / "stanza"
-
-        self.resource_dir_pathlib = pathlib.Path(resource_dir)
-        self.not_entities_path = self.resource_dir_pathlib / "concatenator" / "not_entities.txt"
-        self.space_between_not_ok_path = self.resource_dir_pathlib / "concatenator" / "space_between_not_ok.txt"
-        self.months_path = self.resource_dir_pathlib / "concatenator" / "months.txt"
+        self.custom_ner_model_Path = pathlib.Path(self.resource_dir) / "ner_models"
 
         self.prepare_resources(refresh_data)
 
         self._entity_mapper = None
         self.loaded_entity_files = []
 
-        self.not_entities = self._load_not_entities()
-        self.space_between_not_ok = self._load_space_between_not_ok()
-        self.months = self._load_months()
-        self.concat_resources = {
-            "months": self.months,
-            "space_between_not_ok": self.space_between_not_ok,
-            "not_entities": self.not_entities
-        }
 
 
     def prepare_resources(self, refresh_data):
@@ -115,10 +125,27 @@ class MLP:
             shutil.rmtree(self.resource_dir)
             self.logger.info("MLP data directory deleted.")
         # download resources
+        self.download_custom_ner_models(self.resource_dir, logger=self.logger, model_langs=self.custom_ner_model_langs)
         self.download_stanza_resources(self.resource_dir, self.supported_langs, logger=self.logger)
         self.download_entity_mapper_resources(self.resource_dir, logger=self.logger)
-        self.download_concatenator_resources(self.resource_dir, logger=self.logger)
 
+    @staticmethod
+    def download_custom_ner_models(resource_dir: str, logger=None, custom_ner_model_urls: dict = CUSTOM_NER_MODELS, model_langs: list=None):
+        """
+        Downloads custom ner models if not present in resources directory.
+        """
+        ner_resource_dir = pathlib.Path(resource_dir) / "ner_models"
+        ner_resource_dir.mkdir(parents=True, exist_ok=True)
+        for lang, url in custom_ner_model_urls.items():
+            if lang in model_langs:
+                file_name = urlparse(url).path.split("/")[-1]
+                file_path = ner_resource_dir / lang
+                if not file_path.exists():
+                    if logger: logger.info(f"Downloading custom ner model file {file_name} into directory: {url}")
+                    response = urlopen(url)
+                    content = response.read()
+                    with open(file_path, "wb") as fh:
+                        fh.write(content)
 
     @staticmethod
     def download_stanza_resources(resource_dir: str, supported_langs: List[str], logger=None):
@@ -221,7 +248,6 @@ class MLP:
             json_doc=json_object,
             doc_path=doc_paths,
             entity_mapper=self.get_entity_mapper(),
-            concat_resources=self.concat_resources,
             error=e
         )
         return document
@@ -241,24 +267,44 @@ class MLP:
 
     def get_stanza_pipeline(self, lang: str):
         if lang not in self._stanza_pipelines:
-            try:
-                self._stanza_pipelines[lang] = stanza.Pipeline(
-                    lang=lang,
-                    dir=str(self.stanza_resource_path),
-                    processors=self._get_stanza_processors(lang),
-                    use_gpu=self.use_gpu,
-                    logging_level=self.logging_level,
-                )
-
-            # This is for CUDA OOM exceptions. Fall back to CPU if needed.
-            except RuntimeError:
-                self._stanza_pipelines[lang] = stanza.Pipeline(
-                    lang=lang,
-                    dir=str(self.stanza_resource_path),
-                    processors=self._get_stanza_processors(lang),
-                    use_gpu=False,
-                    logging_level=self.logging_level,
-                )
+            if lang in self.custom_ner_model_langs:
+                try:
+                    self._stanza_pipelines[lang] = stanza.Pipeline(
+                        lang=lang,
+                        dir=str(self.stanza_resource_path),
+                        processors=self._get_stanza_processors(lang),
+                        ner_model_path=f"{self.custom_ner_model_Path}/{lang}",
+                        use_gpu=self.use_gpu,
+                        logging_level=self.logging_level,
+                    )
+                # This is for CUDA OOM exceptions. Fall back to CPU if needed.
+                except RuntimeError:
+                    self._stanza_pipelines[lang] = stanza.Pipeline(
+                        lang=lang,
+                        dir=str(self.stanza_resource_path),
+                        processors=self._get_stanza_processors(lang),
+                        ner_model_path=f"{self.custom_ner_model_Path}/{lang}",
+                        use_gpu=False,
+                        logging_level=self.logging_level,
+                    )
+            else:
+                try:
+                    self._stanza_pipelines[lang] = stanza.Pipeline(
+                        lang=lang,
+                        dir=str(self.stanza_resource_path),
+                        processors=self._get_stanza_processors(lang),
+                        use_gpu=self.use_gpu,
+                        logging_level=self.logging_level,
+                    )
+                # This is for CUDA OOM exceptions. Fall back to CPU if needed.
+                except RuntimeError:
+                    self._stanza_pipelines[lang] = stanza.Pipeline(
+                        lang=lang,
+                        dir=str(self.stanza_resource_path),
+                        processors=self._get_stanza_processors(lang),
+                        use_gpu=False,
+                        logging_level=self.logging_level,
+                    )
 
         return self._stanza_pipelines[lang]
 
@@ -275,6 +321,7 @@ class MLP:
                 for word in sentence.words:
                     words.append(word)
                 sentences.append(words)
+
                 for entity in sentence.entities:
                     entities.append(entity)
 
@@ -288,11 +335,6 @@ class MLP:
         return sentences, entities, e
 
 
-    def _get_stanza_ner(self, lang: str, raw_text: str):
-        pipeline = self.get_stanza_pipeline(lang)(raw_text)
-        return [entity for sentence in pipeline.sentences for entity in sentence.entities]
-
-
     @staticmethod
     def _get_stanza_processors(lang):
         """
@@ -304,7 +346,7 @@ class MLP:
             return "tokenize,pos,lemma"
 
 
-    def process(self, raw_text: str, analyzers: list = ["all"], lang=None):
+    def process(self, raw_text: str, analyzers: list = DEFAULT_ANALYZERS, lang=None, spans: str = "text"):
         """
         Processes raw text.
         :param: raw_text str: Text to be processed.
@@ -319,6 +361,10 @@ class MLP:
                 # For every analyzer, activate the function that processes it from the
                 # document class.
                 self.__apply_analyzer(document, analyzer)
+
+            if "sentences" in analyzers and spans == "sentence":
+                document.fact_spans_to_sent()
+
             return document.to_json()
         else:
             return None
@@ -363,7 +409,7 @@ class MLP:
             self.logger.exception(e)
 
 
-    def process_docs(self, docs: List[dict], doc_paths: List[str], analyzers=["all"]):
+    def process_docs(self, docs: List[dict], doc_paths: List[str], analyzers=DEFAULT_ANALYZERS, spans: str = "text"):
         """
         :param docs: Contains tuples with two dicts inside them, the first being the document to be analyzed and the second is the meta information that corresponds to the document for transport purposes later on.
         :param doc_paths: Dot separated paths for how to traverse the dict for the text value you want to analyze.
@@ -386,6 +432,9 @@ class MLP:
                             # document class.
                             self.__apply_analyzer(doc, analyzer)
 
+                        if "sentences" in analyzers and spans == "sentence":
+                            doc.fact_spans_to_sent()
+
                         result = doc.document_to_json(use_default_doc_path=False)
                         new_facts = result.pop("texta_facts", [])
                         existing_facts = document.get("texta_facts", [])
@@ -404,7 +453,6 @@ class MLP:
 
         return container
 
-
     @staticmethod
     def download_concatenator_resources(resource_dir: str, logger):
         concat_resource_dir = pathlib.Path(resource_dir) / "concatenator"
