@@ -141,7 +141,7 @@ class ContactEmailParser(AbstractParser):
         """
         pattern = re.compile(r'([^\s]+\s?\.\s?)?[^\s@]+\s?@\s?[^@.\s]+\s?[.]\s?[^\s]{1,20}')
         for email in pattern.finditer(self.text):
-            yield ParseResult(value=email.group(0).replace(' ', ''), span=email.span())
+            yield ParseResult(value=email.group(0).replace('mailto:', '').replace(' ', ''), span=email.span())
 
 
 class ContactEmailNamePairParser(AbstractParser):
